{
open Lexing
open Parser

exception SyntaxError of string

let next_line lexbuf =
  let pos = lexbuf.lex_curr_p in
  lexbuf.lex_curr_p <-
    { pos with pos_bol = lexbuf.lex_curr_pos;
               pos_lnum = pos.pos_lnum + 1
    }
}

let white = [' ' '\t']+
let newline = '\r' | '\n' | "\r\n"
let key = 'k' ['a'-'z' 'A'-'Z' '0'-'9' '\'''_']*
let var = ['x' 'y' 'z'] ['a'-'z' 'A'-'Z' '0'-'9' '_' '\'']*
let nonce = ['n' 'm' 'r'] ['a'-'z' 'A'-'Z' '0'-'9' '_' '\'']*
let nonce_star = ['n' 'm' 'r'] ['a'-'z' 'A'-'Z' '0'-'9' '_' '\'']* '*'
let const = ['A' 'B' 'C' 'a' 'b' 'c'] ['a'-'z' 'A'-'Z' '0'-'9' '_' '\'']*
let comment = '(' '*' ([^')'] | [^'*']')') * '*' ')'

rule read =
  parse
  | white    { read lexbuf }
  | comment  { read lexbuf }
(*   | '"'      { read_string (Buffer.create 17) lexbuf } *)
  | '('      { LEFT_PARAN }
  | ')'      { RIGHT_PARAN }
  | '{'      { LEFT_BRACE }
  | '}'      { RIGHT_BRACE }
  | ':'      { COLON }
  | ','      { COMMA }
  | '!'      { EXCLAMATION }
  | '.'      { DOT }
  | '*'      { STAR }
  | '+'      { PLUS }
  | '_'      { UNDERSCORE }
  | '|'      { PIPE }
  | '='      { EQUAL }
  | "~~~"    { EQUIV }
  | "###"    { SEPARATOR }
  | '0'      { ZERO }
  | "->"     { ARROW }
  | "enc"    { ENC }
  | "aenc"   { AENC }
  | "inl"    { INL }
  | "inr"    { INR }
  | "pk"     { PK }
  | "vk"     { VK }
  | "dec"    { DEC }
  | "sign"   { SIGN }
  | "checksign" { VERIFY }
  | "adec"   { ADEC }
  | 'h'      { HASH }
  | "p_1"    { PROJ1 }
  | "p_2"    { PROJ2 }
  | "new"    { NEW }
  | "in"     { IN }
  | 'v'      { OR }
  | "out"    { OUT }
  | "let"    { LET }
  | "else"   { ELSE }
  | "if"     { IF }
  | "then"   { THEN }
  | "match"  { MATCH }
  | "with "  { WITH }
  | 'L'     { LL }
  | 'H'     { LH }
  | 'S'     { LS }
  | "key"   { KEYTYPE }
  | "x_LR"   { XLR }
  | "x_L"   { XL }
  | "x_R"   { XR }
  | eof      { EOF }
  | newline  { next_line lexbuf; read lexbuf }
  | key      { KEY (Lexing.lexeme lexbuf) }
  | var      { VAR (Lexing.lexeme lexbuf) }
  | nonce_star    { NONCE_STAR (Lexing.lexeme lexbuf) }
  | nonce    { NONCE (Lexing.lexeme lexbuf) }
  | const    { CONST (Lexing.lexeme lexbuf) }
  | _ { raise (SyntaxError ("Unexpected char: " ^ Lexing.lexeme lexbuf)) }
