open Syntax
open Util



(* reduce destructor applications *)
let eval (d:dest) (t:term) : term option = 
  match d with 
  | DSDec (_,x, k) -> 
    (match t with 
    | VSEnc (_,t', k') -> 
      if k = k' then Some t' else None
    | _ -> None)
  | DADec (_,x, k) -> 
    (match t with 
    | VAEnc (_,t', PKey k') -> 
      if k = k' then Some t' else None
    | _ -> None)
  | DProj1 (_,p) -> 
    (match t with 
    | VPair (_,t1, t2) -> Some t1
    | _ -> None)
  | DProj2 (_,p) -> 
    (match t with 
    | VPair (_,t1, t2) -> Some t2
    | _ -> None)
  | DVer (_,x, VerKey k) -> 
    (match t with 
    | VSign (_,t', k') -> 
      if k = k' then Some t' else None
    | _ -> None)

(* Exception thrown when encountering a typing error *)
exception Type_Error of string * error * error

let rec subtyping (t1:ty) (t2:ty) = 
  (* SRefl *)
  same_type t1 t2 ||
  match t1, t2 with 
  (* SHigh *)
  | _ , TLabel LH -> true
  (* SPair *)
  | TPair (t1', t1''), TPair (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SPairL *)
  | TPair (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SPairS  + SPairS'*)
  | TPair (t1', t1''), TLabel LS -> subtyping t1' (TLabel LS) ||  subtyping t1'' (TLabel LS)
  (* SSum *)
  | TSum (t1', t1''), TSum (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SSumL *)
  | TSum (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SKey *)
  | TKey (l,t), TLabel l' -> l = l'
  (* SEnc *)
  | TSEnc (t,k), TSEnc (t',k') -> k = k' &&  subtyping t t'
  (* SAEnc *)
  | TAEnc (t,k), TAEnc (t',k') -> k = k' &&  subtyping t t'
  (* Rule STrans is incomporated directly in the other rules *)
  | _ -> false
   

let rec tct_new (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) (t:ty) : (bool * constr_set) =  
  let b,c = 
    try 
    (

    match u, v, t with 
    (* THigh *)
    | _ ,_ , TLabel LH -> true, []

    (* TOr *)
    | u, v, TOr (t1, t2) -> 
      let b1, c1 = tct_new e e' g d u v t1 in
      let b2, c2 = tct_new e e' g d u v t2 in 
      if b1 then
        true, c1
      else
        b2, c2

    (* TNonce + TNonceL *)
    | VNonce (er,n), VNonce (er',m), TLabel l ->
      (match e n, e' m with 
       | Some l1 (*(l1, _)*), Some l2 (*(l2, _)*) -> 
        if l1 = l2 && l1 = l && (l1 <> LL || n = m) then 
          true, []
        else
          raise (Type_Error ("Types of the nonces in the environment don't match. Type of " ^ nonce2string n ^ " is " ^ label2string l1 ^ " ,Type of " ^ nonce2string m ^ " is " ^ label2string l2, er, er'))
      | None, None -> 
        if n = m then 
          if l = LL then
            true, []
          else
            raise (Type_Error ("Unbound nonce is always Low",er,er'))
        else raise (Type_Error ("Nonce not bound on both sides and not equal", er, er'))
      | None, _ -> raise (Type_Error ("Nonce " ^ nonce2string n ^ " is not bound (on the left)", er, er'))
      | _ -> raise (Type_Error ("Nonce " ^ nonce2string m ^ " is not bound (on the right)", er, er')))

    (* TPair *)
    | VPair (er,n, m), VPair (er',n', m'), TPair (t1, t2) -> 
      let b1,c1 = tct_new e e' g d n n' t1 in
      let b2,c2 = tct_new e e' g d m m' t2 in
      b1 && b2, (c1 @ c2) 

    (* TConst *)
    | VConst (er,c), VConst (er',c'), TLabel LL -> 
      if c = c' then   
        true, []
      else
        raise (Type_Error ("Constants " ^ c ^ " and " ^ c' ^ " are not equal", er, er'))

    (* TInl *)
    | VInl (er,n), VInl (er',m), TSum (t1, t2) ->
      tct_new e e' g d n m t1

    (* TInr *)
    | VInr (er,n), VInr (er',m), TSum (t1, t2) -> 
      tct_new e e' g d n m t2

    (* TKey *)
    | VKey (er, k), VKey (er', k'), TKey (l, t) -> 
      if k = k' then
        match d k with 
        | Some t' -> TKey (l,t) = t', []
        | None -> raise (Type_Error ("key " ^ k ^ " is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TPubKey *)
    | VPKey (er, PKey k), VPKey (er', PKey k'), TLabel LL -> 
      if k = k' then
        if d k <> None then true, []
        else raise (Type_Error ("Public key pk(" ^ k ^") is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TVKey *)
    | VVerKey (er, VerKey k), VVerKey (er', VerKey k'), TLabel LL -> 
      if k = k' then
        if d k <> None then true, []
        else raise (Type_Error ("Verification key vk(" ^ k ^") is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TVar *)
    | VVar (er,x), VVar (er',y), t -> 
      (match lookup_var g x, lookup_var g y, t with
        (* Instantiation of variables as a specialized version of TLR' combined with TLRVar *)
        | Some (TRefine (u1,v1)), Some (TRefine (u2,v2)), _ -> 
          if not (is_star_nonce u1 || is_star_nonce u2 || is_star_nonce v1 || is_star_nonce v2) then 
            tct_new e e' g d u1 v2 t
          else
            (* TLR'* *)
            if (is_star_nonce u1 && is_star_nonce u2 && is_star_nonce v1 && is_star_nonce v2 && x = y) then 
              tct_new e e' g d u1 v2 t
            else
             raise (Type_Error ("cannot instantiate ill-formed refinements", er, er'))
        | Some (TSEnc (t1,k1)), Some (TSEnc (t2,k2)), TLabel LL ->
            if x = y then
              (match d k1 with 
              | Some (TKey (LS, t)) -> 
                  subtyping t1 t,[(u,v)] 
              | Some (TKey (LL, t)) -> true, []
              | _ -> raise (Type_Error ("Key " ^ k1 ^ " is not bound or key is H", er, er')))
            else
              raise (Type_Error ("variables " ^ x ^ " and " ^ y ^ " are not equal", er, er'))
        | Some (TAEnc (t1,k1)), Some (TAEnc (t2,k2)), TLabel LL ->
            if x = y then
              (match d k1 with 
              | Some (TKey (LS, t)) -> 
                subtyping t1 t, [(u,v)] 
              | Some _ -> true, []
              | _ -> raise (Type_Error ("Key " ^ k1 ^ " is not bound or key is H", er, er')))
            else
              raise (Type_Error ("variables " ^ x ^ " and " ^ y ^ " are not equal", er, er'))
        | _ -> 
        (if x = y then
          match lookup_var g x with 
          | Some (TOr _)  ->  raise (Type_Error ("Variables may not be output with or type", er, er'))
          | Some t' -> subtyping t' t, []
          | None -> raise (Type_Error ("variable " ^ var2string x ^ " is not bound", er, er'))
        else
          raise (Type_Error ("variables " ^ var2string x ^ " and " ^ var2string y ^ " are not equal", er, er'))))

    (* TEnc *)
    | VSEnc (er, n, k), VSEnc (er',m, k'), TSEnc (t,k'') ->
      if k = k' && k = k'' then
        tct_new e e' g d n m t
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TEncH + TEncL *)
    | VSEnc (er, n, k), VSEnc (er',m, k'), TLabel LL ->
      if k = k' then
        (match d k with 
        | Some (TKey (LS, t)) -> 
          let b,c = tct_new e e' g d u v (TSEnc (t,k)) in 
          b, (u,v)::c 
        | Some (TKey (LL, t)) -> 
          tct_new e e' g d u v (TSEnc (TLabel LL,k))
        | _ -> 
            raise (Type_Error ("Key " ^ k ^ " is not bound or key is H", er, er')))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TAEnc *)
    | VAEnc (er, n, PKey k), VAEnc (er',m, PKey k'), TAEnc (t,k'') ->
      if k = k' && k = k'' then
        tct_new e e' g d n m t
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TAEncH + TAEncL *)
    | VAEnc (er, n, PKey k), VAEnc (er',m, PKey k'), TLabel LL ->
      if k = k' then
        (match d k with 
        | Some (TKey (LS, t)) -> 
            let b, c = tct_new e e' g d u v (TAEnc (t,k)) in 
            b, (u,v) :: c
        | Some _ -> 
          tct_new e e' g d u v (TAEnc (TLabel LL,k))
        | _ -> 
            raise (Type_Error ("Key " ^ k ^ " is not bound", er, er')))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TSignH + TSignL*)
    | VSign (er,n, k), VSign (er',m, k'), TLabel LL ->
      if k = k' then
        match d k with 
        | Some (TKey (LS, t)) -> 
            let b1, c1 = tct_new  e e' g d n m t in 
            let b2, c2 = tct_new  e e' g d n m (TLabel LL) in 
            b1 && b2, (u,v) :: c1 @ c2
        | Some (TKey (LL, t')) -> 
            tct_new  e e' g d n m (TLabel LL) 
        | _ -> raise (Type_Error ("Incompatible types of signing key and payload", er, er')) else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* THash + THashL *)
    | VHash (er,n), VHash (er',m), TLabel LL -> 
        (* Heuristic: we use THash if hash contains an S term directly under pair *)
        if s_under_pair e d n && s_under_pair e' d m then
          true, [u,v]
        else
          tct_new e e' g d n m (TLabel LL)

    (* TLR *)
    | u, v, TRefine (tl, tr) -> 
        print_verbose 2 ("Refinment Check! " ^  term2string u ^ ", " ^ term2string v ^ " : " ^ term2string tl ^ ", " ^ term2string tr);
        if ground u && ground v then
          if well_formed_refinement u v then
            same_term u tl && same_term v tr, []
          else
            raise (Type_Error ("Refinement Type is not well formed", term_error u, term_error v))
        else
          raise (Type_Error ("Terms for refinement types must be ground",term_error u, term_error v))



    (***** TSub Rules *****)

    (* SHigh not needed *)
    (* SRefl not needed *)
    (* STrans implicitly by recursion *)
    (* SPair not needed *)

    (* SPairL *)
    | VPair (er,n, m), VPair (er',n', m'), TLabel LL -> 
      tct_new e e' g d u v (TPair (TLabel LL,TLabel LL)) 

    (* SPairS *)
    | VPair (er,n, m), VPair (er',n', m'), TLabel LS -> 
      let b1, c1 = tct_new e e' g d u v (TPair (TLabel LS,TLabel LH)) in 
      if b1 then b1, c1 else
      tct_new e e' g d u v (TPair (TLabel LH,TLabel LS)) 

    (* STU not needed *)

    (* STUL *)
    | VInl (er,n), VInl (er',n'), TLabel LL -> 
      tct_new e e' g d u v (TSum (TLabel LL,TLabel LL)) 
    | VInr (er,n), VInr (er',n'), TLabel LL -> 
      tct_new e e' g d u v (TSum (TLabel LL,TLabel LL)) 

    (* SKey *)
    | VKey (er, k), VKey (er',k'), TLabel l ->
      if k = k' then
        match d k with 
        | Some (TKey (l', t)) -> l = l', []
        | _ -> raise (Type_Error ("key " ^ k ^ " is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))


    | _ -> raise (Type_Error ("Term Typing Failure : " ^ term2string u ^ " ~ " ^ term2string v ^ " /:/ " ^ type2string t, term_error u, term_error v))     
    )
  with
    | Type_Error (s,er,er') -> print_verbose 2 ("WARNING!\n"  ^ s); false, [] 

  in 
  print_verbose 2 ("Typing term : " ^ term2string u ^ ", " ^ term2string v ^ " : " ^ type2string t ^  " returns " ^ string_of_bool b); 
  if c <> [] then 
    print_verbose 2 ("Got non-empty constraints"); 
  b,c




let tcd (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (de:dest) (de':dest): ty =  
  match de,de' with 
  | DSDec (er,x, k), DSDec (er',x2,k2) -> 
    if x = x2 && k = k2 then
      (match d k, lookup_var g x with
      (* DDecH *)
      | Some (TKey (LS, t)), Some (TLabel LL) -> t
      (* DDecL *)
      | Some (TKey (LL, t)), Some (TLabel LL) -> TLabel LL
      (* DDecT *)
      | _ , Some (TSEnc (t,k')) -> t
      | Some t, Some t' -> raise (Type_Error ("Symetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Symetric decryption with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable or key is differing",er,er'))

  | DADec (er,x, k), DADec (er',x2,k2) -> 
    if x = x2 && k = k2 then
      (match d k, lookup_var g x with
      | Some (TKey (LS, t)), Some (TLabel LL) -> TOr (t, TLabel LL)
      | Some (TKey (LL, t)), Some (TLabel LL) -> TLabel LL
      | _ , Some (TAEnc (t,k')) -> t
      | Some t, Some t' -> raise (Type_Error ("Assymetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Assymetric decryption with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable or key is differing",er,er'))

  | DVer (er,x, VerKey k), DVer(er',x2,VerKey k2) -> 
    if x = x2 && k = k2 then
    (match d k, lookup_var g x with
      | Some (TKey (LS, t)), Some (TLabel LL) -> t
      | Some (TKey (LL, t)), Some (TLabel LL) -> TLabel LL
      | Some t, Some t' -> raise (Type_Error ("Verification of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Verification with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable or key is differing",er,er'))

  | DProj1 (er,x), DProj1 (er',x2) -> 
    if x = x2 then
      (match lookup_var g x with 
      | Some (TPair (t1, t2)) ->  t1
      | Some (TLabel LL) -> TLabel LL
      | _ -> raise (Type_Error ("Projection not applicable",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | DProj2 (er,x), DProj2 (er',x2) -> 
    if x = x2 then
      (match lookup_var g x with 
      | Some (TPair (t1, t2)) -> t2
      | Some (TLabel LL) ->  TLabel LL
      | _ -> raise (Type_Error ("Projection not applicable",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | _ -> raise (Type_Error ("Destructors do not match",ErTODO,ErTODO))


(* collect all variables with or types  that appear at top level of the
 * processes *)
let collect_top_or (g:var_env) (p:process) (q:process) : var list = 
  let vars = 
    match p, q with 
    | POut (_,u, p'), POut (_,v, q') -> 
      let v_u = snd (collectnv_t u) in 
      let v_v = snd (collectnv_t v) in 
      v_u @ v_v
    | PLet (_,x, de, p', p''), PLet (_,x', de', q', q'') ->
      [dest_var de; dest_var de']
    | PIf (_,u, v, p', p''), PIf (_,u', v', q', q'') -> 
      let v_u = snd (collectnv_t u) in 
      let v_u' = snd (collectnv_t u') in 
      let v_v = snd (collectnv_t v) in 
      let v_v' = snd (collectnv_t v') in 
      v_u @ v_u' @ v_v @ v_v'
    | PMatch (_,x, y, p', z, p''), PMatch (_,x', y', q', z', q'') -> 
      [x;x']
    | _ -> []
  in List.sort_uniq Pervasives.compare 
    (List.filter (fun x -> match lookup_var g x with Some TOr _ -> true | _ -> false) vars)


(* This represents a weaker form of instantiation *)
let rec infer_refinement (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) : ty option = 
  match u,v with 
  | VVar (_,x), VVar (_,y) ->
      (match lookup_var g x, lookup_var g y with
      | Some (TRefine (u1,u2)), Some (TRefine (v1, v2)) -> 
        if not (is_star_nonce u1 || is_star_nonce u2 || is_star_nonce v1 || is_star_nonce v2) || x = y then
          Some (TRefine (u1, v2))
        else 
          None
      | _ -> None)
  | _ -> if  well_formed_refinement u v then 
           Some (TRefine (u, v))
         else
           None



let incompatible_types (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (u':term) (v:term) (v':term) : bool = 
  (* PIfS*)
  let b1, _ = tct_new e e' g d u u' (TLabel LL) in
  let b2, _ = tct_new e e' g d v v' (TLabel LS) in
  if b1 && b2 then
    true
  else
    (* PIfS symmetric version *)
    let b1, _ = tct_new e e' g d u u' (TLabel LS) in
    let b2, _ = tct_new e e' g d v v' (TLabel LL) in
    if b1 && b2 then 
      true
    else
      (* PIfI *)
      match (infer_refinement e e' g d v v') with
      | Some (TRefine (v1, v2)) ->
          let pairs2 = match v1, v2 with 
            | VPair _, _ -> true
            | _ , VPair _ -> true
            | _ -> print_verbose 2 "none are pairs"; false  in 
          let pairs1 = fst (tct_new e e' g d u u' (TPair (TLabel LH, TLabel LH)))
          in pairs1 && not pairs2
      | _ -> false

(* Typing rules for processes *)
let rec tcp (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (p:process) (q:process) : constr_set_set =
  (* This part takes care of rule TOr *)
  let top_or = collect_top_or g p q in 
  if top_or <> [] then
    let x = List.hd top_or in 
    let (t1,t2) = 
      match lookup_var g x with 
      | Some (TOr (t1,t2)) -> t1,t2 
      | _ -> failwith "This has to be an or type"
    in (tcp e e' (extend_var_env g x t1) d p q) @ (tcp e e' (extend_var_env g x t2) d p q)

  else
    match p, q with 
    (* PZero *)
    | PZero er , PZero er'-> 
        [([], g)] (* Branching is done in a final step *)

    (* POut *)
    | POut (er, u, p'), POut (er', v, q') ->
      let b,c = tct_new e e' g d u v (TLabel LL) in 
      if b then
        union_forall c (tcp e e' g d p' q') 
      else 
        raise (Type_Error ("Output message " ^ term2string u ^ " ~ " ^ term2string v ^ " cannot be typed with type L."(* ^ "Inferred type is " ^ type2string (fst (tct_real e e' g d u v))*),er,er'))

    (* PIn *)
    | PIn (er, x, p'), PIn (er', y, q') ->
      if x = y then
        if lookup_var g x = None then
          let css = tcp e e' (extend_var_env_nou g x (TLabel LL)) d p' q' in
          clean_constr_set_set_var css x
        else 
          raise (Type_Error ("Variable " ^ var2string x ^ " used in in was already bound",er,er'))
      else 
        raise (Type_Error ("Variables " ^ var2string x ^ " and " ^ var2string y ^ " used in in are not equal ",er,er'))

    (* PNew *)
    | PNew (er, n, l, p'), q -> 
       if e n = Some l  then 
         tcp e e' g d p' q
       else 
         raise (Type_Error ("Annotated nonce type for " ^ nonce2string n ^ " does not match environment type",er,er))

    (* PNew' *)
    | p, PNew (er, n, l, q') -> 
       if e' n = Some l then 
         tcp e e'  g d p q'
      else 
        raise (Type_Error ("Annotated nonce type for " ^ nonce2string n ^ " does not match environment type",er,er))

    (* PPar *)
    | PPar (er, p', p''), PPar (er', q', q'') ->
      add_cart (tcp e e' g d p' q') (tcp e e' g d p'' q'')

    | PRepl (er, p'), PRepl (er', q') -> failwith "Replication should not occure when typing."

    (* PLet + PLetLR + PLetLR* *)
    | PLet (er, x, de, p', p''), PLet (er', x', de', q', q'') ->
       if same_de de de' then 
         if lookup_var g x = None then 
           (match lookup_var g (dest_var de) with
            | Some (TRefine (m,n)) -> 
               if ground n && ground m then  
                 (* PLetLR* *)
                 if is_star_nonce n && is_star_nonce m then
                   tcp e e' g d p'' q''
                 (* PLetLR *)
                 else
                   (match eval de m, eval de n with
                    | None, None ->
                       tcp e e' g d p'' q'' 
                    | _ -> raise (Type_Error ("Not handled yet, and should not occur with atomic refinements", er, er')))
               else
                 raise (Type_Error ("Refinement in Let is not ground", er ,er'))
            (* PLet *)
            | Some tv -> 
               let t = tcd e e' g d de de' in 
            
               let css = tcp e e' (extend_var_env_nou g x t)  d p' q' in
               clean_constr_set_set_var css x @ (tcp e e' g d p'' q'')
            | None -> raise (Type_Error ("Variable " ^ var2string (dest_var de) ^ " used in let is not bound",er,er')))
         else
           raise (Type_Error ("Variable "  ^ var2string x ^ " used in let is already bound",er,er'))
       else
         raise (Type_Error ("Structure of the two let expressions is different",er,er'))

    (* PIf *)
    | PIf (er, u, v, p', p''), PIf (er', u', v', q', q'') ->
       (match infer_refinement e e' g d u u', infer_refinement e e' g d v v' with
       | Some (TRefine (u1,u2)), Some (TRefine (v1,v2)) ->
          (* PIfLR* *)
          if is_star_nonce u1 && is_star_nonce u2 && is_star_nonce v1 && is_star_nonce v2 && same_term u1 u2 && same_term v1 v2 then
            tcp e e' g d p' q' @ tcp e e' g d p'' q''
          else
            (* PIfLR'* *)
            if not (is_star_nonce u1) && not (is_star_nonce u2) && is_star_nonce v1 && is_star_nonce v2 then
              tcp e e' g d p' q' @ tcp e e' g d p'' q'' 
            (* symmetric variant of PIfLR'* *)
            else
              if not (is_nonce v1) && not (is_nonce v2) && is_star_nonce u1 && is_star_nonce u2 then
                tcp e e' g d p' q' @ tcp e e' g d p'' q'' 
              (* PIfLR *)
              else
                (match same_term u1 v1, same_term u2 v2 with 
                  | true, true -> tcp e e' g d p' q'
                  | true, false -> tcp e e' g d p' q''
                  | false, true -> tcp e e' g d p'' q'
                  | false, false -> tcp e e' g d p'' q'')
       | _ -> 
          (* PIfS +_PIf + PIf* *) 
          if incompatible_types e e' g d u u' v v'  then
            tcp e e' g d p'' q'' 
          else
            (* PIfL *)
            let bt1, ct1 = tct_new e e' g d u u' (TLabel LL) in
            let bt2, ct2 = tct_new e e' g d v v' (TLabel LL) in
            if bt1 && bt2 then 
              let cp1 = tcp e e' g d p' q' in 
              let cp2 = tcp e e' g d p'' q'' in 
              (* PIfP *)
              let pifp = 
                (match v, v' with 
                 | VVar (_,t), VVar (_,t') -> t = t'
                 | VConst (_,t), VConst (_,t') -> t = t'
                 | VNonce (_,t), VNonce (_,t') -> t = t'
                 | _ -> false) in 
              if pifp then
                cp1 @ cp2
              else
                union_forall (ct1 @ ct2) (cp1 @ cp2)
            else
              raise (Type_Error ("Illegal constellation for IF: \n",er,er')))
    | PMatch (er, x, y, p', z, p''), PMatch (er', x', y', q', z', q'') -> 
       if x = x' && y = y' && z = z' then
         (match lookup_var g x with 
          | Some (TSum (t, t')) -> 
             let css = tcp e e' (extend_var_env g y t)  d p' q' in
             let css' = tcp e e' (extend_var_env g z t') d p'' q'' in
             (clean_constr_set_set_var css y) @ (clean_constr_set_set_var css' z) 

          | Some t -> 
             if t = (TLabel LL) then
               let css = tcp e e' (extend_var_env g y (TLabel LL))  d p' q' in
               let css' = tcp e e' (extend_var_env g z (TLabel LL)) d p'' q'' in
               (clean_constr_set_set_var css y) @ (clean_constr_set_set_var css' z)
             else
               raise (Type_Error ("Variable " ^ var2string x ^ " used in Match does not have type L",er,er'))
          | None -> raise (Type_Error ("Variable " ^ var2string x ^ " used in Match is not bound",er,er')))
       else
         raise (Type_Error ("Variables of the two match expressions are different",er,er'))
      
    | _ -> raise (Type_Error  ("Case not covered by typing rules",process_error p ,process_error q))
     


let collectnv_cs ((cs):constr_set) : nonce list * var list = 
  List.fold_left 
  (fun (nl,vl) c -> 
    let ns,vs = collectnv_c c in 
    List.sort_uniq Pervasives.compare (nl @ ns), 
    List.sort_uniq Pervasives.compare (vl @ vs)) 
  ([],[]) cs

(* We implement branching only for the variables that appear in the
 * constraints, not for all *)
let branching (cs:constr_set_env) : constr_set_set = let g = snd cs in 
  let v_or =  
    List.filter (fun x -> match lookup_var g x with Some (TOr (t1, t2)) -> true | _ -> false) 
      (snd (collectnv_cs (fst cs))) in 
  List.fold_left 
    (fun css x -> match lookup_var g x with 
      | Some (TOr (t1,t2)) -> 
        List.map (fun (cs,g') -> (cs, extend_var_env g' x t1)) css  @ 
        List.map (fun (cs,g') -> (cs, extend_var_env g' x t2)) css
      | _ -> failwith "List should only contain Or types"
      )
    [cs] v_or


let branching_css (css:constr_set_set) : constr_set_set = 
  List.flatten (List.map branching css) 

